import type {
	APIAttachment,
	APIGuildMember,
	APIInteractionDataResolvedChannel,
	APIInteractionGuildMember,
	APIInteractionResponse,
	APIInteractionResponseChannelMessageWithSource,
	APIInteractionResponseDeferredChannelMessageWithSource,
	APIMessage,
	APIModalInteractionResponse,
	APIRole,
	APIUser
} from 'discord-api-types/v10';

export type Handler<
	T extends ApplicationCommandInteractionBase | Error,
	R extends APIInteractionResponse = APIInteractionResponse
> = (interaction: T) => R | Promise<R> | Response | Promise<Response>;

export type ApplicationCommandResponse =
	| APIInteractionResponseChannelMessageWithSource
	| APIInteractionResponseDeferredChannelMessageWithSource
	| APIModalInteractionResponse;

export type SlashCommandHandler = Handler<SlashCommandInteraction, ApplicationCommandResponse>;
export type UserCommandHandler = Handler<UserCommandInteraction, ApplicationCommandResponse>;
export type MessageCommandHandler = Handler<MessageCommandInteraction, ApplicationCommandResponse>;
export type ErrorHandler = Handler<Error, ApplicationCommandResponse>;

export type InteractionHandler = {
	slash?: SlashCommandHandler;
	user?: UserCommandHandler;
	message?: MessageCommandHandler;
	error?: ErrorHandler;
};

export type ApplicationCommandInteractionBase = {
	id: string;
	name: string;
	application_id: string;
	channel_id: string;
	command_id: string;
	guild_id: string | null;
	guild_locale: string | null;
	app_permissions: string | null;
	commandType: number;
	member: APIInteractionGuildMember | null;
	user: APIUser | null;
	token: string;
	fullName: string;
	type: 2;
};

export type SlashCommandInteraction = Omit<ApplicationCommandInteractionBase, 'commandType'> &
	OptionalSubcommandGroups & { options: Options; commandType: 1 };
export type UserCommandInteraction = Omit<ApplicationCommandInteractionBase, 'commandType'> & {
	data: {
		user: APIUser;
		member: APIInteractionMemberWithUser | null;
	};
	commandType: 2;
};
export type MessageCommandInteraction = Omit<ApplicationCommandInteractionBase, 'commandType'> & {
	data: {
		message: APIMessage;
	};
	commandType: 3;
};

export type ApplicationCommandInteraction =
	| SlashCommandInteraction
	| MessageCommandInteraction
	| UserCommandInteraction;

export type OptionalSubcommandGroups =
	| {
			subcommand: null;
			subcommandGroup: null;
	  }
	| {
			subcommand: string;
			subcommandGroup: null;
	  }
	| {
			subcommand: string;
			subcommandGroup: string;
	  };

export type OptionString = {
	type: ApplicationCommandOptionType.String;
	value: string;
};

export type OptionNumber = {
	type: ApplicationCommandOptionType.Integer | ApplicationCommandOptionType.Number;
	value: number;
};

export type OptionBoolean = {
	type: ApplicationCommandOptionType.Boolean;
	value: boolean;
};

export type OptionUser = {
	type: ApplicationCommandOptionType.User;
	value: {
		user: APIUser;
		member: APIInteractionMemberWithUser | null;
	};
};

export type OptionChannel = {
	type: ApplicationCommandOptionType.Channel;
	value: APIInteractionDataResolvedChannel;
};

export type OptionRole = {
	type: ApplicationCommandOptionType.Role;
	value: APIRole;
};

export type OptionMentionable = {
	type: ApplicationCommandOptionType.Mentionable;
	value: {
		user: APIUser | null;
		member: APIInteractionMemberWithUser | null;
		role: APIRole | null;
	};
};

export type OptionAttachment = {
	type: ApplicationCommandOptionType.Attachment;
	value: APIAttachment;
};

export type Option =
	| OptionString
	| OptionNumber
	| OptionBoolean
	| OptionUser
	| OptionChannel
	| OptionRole
	| OptionMentionable
	| OptionAttachment;
export type Options = Record<string, Option>;

export declare enum ApplicationCommandOptionType {
	Subcommand = 1,
	SubcommandGroup = 2,
	String = 3,
	Integer = 4,
	Boolean = 5,
	User = 6,
	Channel = 7,
	Role = 8,
	Mentionable = 9,
	Number = 10,
	Attachment = 11
}

export type APIInteractionMemberWithUser = Omit<APIGuildMember, 'deaf' | 'mute'>;
